if 0 {
Copyright 2004, 2005 George Peter Staplin  

The following terms apply to all files associated with the software
unless explicitly disclaimed in individual files.

The authors hereby grant permission to use, copy, modify, distribute,
and license this software and its documentation for any purpose, provided
that existing copyright notices are retained in all copies and that this
notice is included verbatim in any distributions. No written agreement,
license, or royalty fee is required for any of the authorized uses.
Modifications to this software may be copyrighted by their authors
and need not follow the licensing terms described here, provided that
the new terms are clearly indicated on the first page of each file where
they apply.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

}

package require Tcl 8.4
proc build.targets {} {


 if {$::argc > 1} {
  foreach t [lrange $::argv 1 end] {
   target:$t
  }
  # We built the targets above, so we are done.
  exit 0
 } 

 foreach t $::targets {
  target:$t
 }}

proc create.target {t deps body} {

 #puts "CREATE TARGET: $t $deps $body"

 proc target:$t {} [string map [list TARGET $t DEPS $deps BODY $body] {
  foreach dep {DEPS} {
   # Build all dependencies
   if {{TARGET} eq $dep} {
    continue
   }

   #puts "should build $dep"
   if {[lsearch -exact $::targets $dep] >= 0} {
    #puts "building $dep"
    target:$dep
   }

   if {[lsearch -exact $::optional $dep] >= 0} {
    target:$dep
   }
  }

  set do_build 0
  foreach dep {DEPS} {
   if {[file.newer? $dep than {TARGET}]} {
    #puts "$dep newer than TARGET"
    # Build this target, because the dependency is newer.
    set do_build 1
   }
  }

  if {$do_build || (![llength {DEPS}])} {
   puts {BUILDING:TARGET}
   namespace eval :: {BODY}
  }
 }]}

proc default.targets {args} {

 set ::targets $args}

proc file.newer? {f _than_ f2} {

 if {![file exists $f] || ![file exists $f2]} {
  return 1
 }
 if {[file mtime $f] > [file mtime $f2]} {
  return 1
 } 
 return 0}

proc main {argc argv} {

 set ::targets [list]
 set ::optional [list]
 set ::building [list]

 uplevel #0 [list source tickmakefile]

 return 0}

proc make {t _from_ deps _using_ body} {

 lappend ::targets $t
 create.target $t $deps $body}

proc meval {args} {

 puts [join $args]
 uplevel #0 $args}

proc mexec {args} {

 set cmd [join $args]
 puts $cmd
 if {[catch {eval exec -- $cmd} msg]} {
  if {"NONE" ne $::errorCode} {
   puts stderr ERROR:$msg
   exit 1
  }
  puts stderr $msg
 }
}

proc need {var} {

 
 switch -- $var {
  CC {
   if {[info exists ::env(CC)]} {
    set ::CC $::env(CC)
   } else {
    set ::CC cc
   }
  }

  CFLAGS {
   if {[info exists ::env(CFLAGS)]} {
    set ::CFLAGS $::env(CFLAGS)
   } else {
    set ::CFLAGS -Wall
   }
  }
 }}

proc optionally.make {t _from_ deps _using_ body} {

 create.target $t $deps $body

 lappend ::optional $t}

proc select.build.conf {} {

 if {0 == $::argc} {
  source ./build.conf
 } elseif {$::argc >= 1} {
  set conf [lindex $::argv 0]
  source $conf
 }}


exit [main $::argc $::argv]
