#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/extensions/record.h>
#include <X11/keysym.h>

struct timeval startTime;
struct timeval thisTime;

volatile sig_atomic_t doExit = 0;

void SetTime (void) {
 if (0 != gettimeofday (&startTime, NULL)) {
  perror ("unable to get time");
 }
}

void GetTime (void) {
 if (0 != gettimeofday (&thisTime, NULL)) {
  perror ("unable to get time");
 }
}

void OutputDelay (FILE *fp) {
 long secDiff;
 long usecDiff;
 long final;

 GetTime ();

 secDiff = (thisTime.tv_sec - startTime.tv_sec);
 usecDiff = ((thisTime.tv_usec - startTime.tv_usec) / 1000);

 /*
  * There are 1000 microseconds in a second.
  */
 final = ((secDiff * 1000) + usecDiff);

 fprintf (fp, "after %ld\n", final);

 SetTime ();
}

void OutputKey (FILE *fp, char *prefix, KeySym ks) {
 if (NoSymbol == ks) {
  perror ("bad keycode");
  return;
 } else if (XK_Pause == ks) {
  exit (EXIT_SUCCESS);
 }

 fprintf (fp, "%s %ld ;# %c\n", prefix, ks, (char) ks);
}

void callback (XPointer p, XRecordInterceptData *idata) {
 Display *dis = (Display *) p;

 if (doExit) {
  exit (EXIT_SUCCESS);
 }

 if (XRecordFromServer == idata->category) {
  xEvent *xev = (xEvent *) idata->data;
  int type = xev->u.u.type;
  char *prefix = "TkXext.send.key.release";

  OutputDelay (stdout);

  switch (type) {
   case ButtonPress: 
    printf ("TkXext.send.button.press %d\n", xev->u.u.detail);
   break;
  
   case ButtonRelease:
    printf ("TkXext.send.button.release %d\n", xev->u.u.detail);
   break;

   case KeyPress:
    prefix = "TkXext.send.key.press";
   case KeyRelease:
   {
    KeyCode kc = xev->u.u.detail;
    KeySym ks = XKeycodeToKeysym (dis, kc, 0);
    OutputKey (stdout, prefix, ks);
   }
   break;

   case EnterNotify:
   case LeaveNotify:
   break;
   
   case MotionNotify:
   {
    int x = xev->u.keyButtonPointer.rootX;
    int y = xev->u.keyButtonPointer.rootY;
    printf ("TkXext.move.pointer %d %d\n", x, y);
   }
   break;
  
   default:
    fprintf (stderr, "INTERNAL TKXEXT ERROR UNKNOWN TYPE (protocol range incorrect?)\n");
    exit (EXIT_FAILURE);
  }
 }

 XRecordFreeData (idata);
}

void sigint_handler (int i) {
 doExit = 1;

 /*
  * Some systems remove the signal handler after it has been called once.
  * In theory it could take a while for the queue to be processed, so we 
  * reset it just in case.
  */
 signal (SIGINT, sigint_handler);
}

int main (int argc, char *argv[]) {
 Display *dis;
 XRecordClientSpec rcspec;
 XRecordRange *xrr;
 XRecordContext rcon;
 Window w;
 Display *otherDis;
 
 signal (SIGINT, sigint_handler);

 SetTime ();

 dis = XOpenDisplay (NULL);

 if (NULL == dis) {
  perror ("unable to open display connection");
  exit (EXIT_FAILURE);
 }

 /*If this is true my X server crashes.*/
 XSynchronize(dis, 0);
 {
  int major, minor;
  XRecordQueryVersion (dis, &major, &minor);
 }

 w = DefaultRootWindow (dis);

 rcspec = XRecordAllClients;
 xrr = XRecordAllocRange ();
 
 if (NULL == xrr) {
  perror ("range allocation failed");
  exit (EXIT_FAILURE);
 }

 xrr->device_events.first = KeyPress;
 xrr->device_events.last = MotionNotify;

 rcon = XRecordCreateContext (dis, 0, &rcspec, 1, &xrr, 1);

 otherDis = XOpenDisplay (NULL);
 if (NULL == otherDis) {
  perror ("unable to open other display connection");
  exit (EXIT_FAILURE);
 }

 XFlush (dis);
 XFlush (otherDis);

 if (!XRecordEnableContext(otherDis, rcon, callback, (XPointer) dis)) {
  perror ("enable context failed");
  exit (EXIT_FAILURE);
 }

 return EXIT_SUCCESS;
}
