#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xmu/WinUtil.h>
#include <tcl.h>
#include <tk.h>
#include "TkXext.h"

#define SEARCH_N_SECONDS_FOR_WINDOW 9
 
static int windowMatchTitle (Display *dis, Window w, char *titleToFind) {
 XTextProperty xtp;
 char *title;
 Window r = None;

 if (!XGetWMName (dis, w, &xtp)) {
   return None;
 }

 if (xtp.value) {
  title = (char *) xtp.value;
  if (Tcl_StringMatch (title, titleToFind)) {
   r = w;
  }
  XFree (title);
 }
 return r;
}

static Window searchWindowTree (Display *dis, Window w, char *titleToFind, Tcl_Time *startTime) {
 Tcl_Time thisTime;
 Window dummy, *children;
 unsigned int nChildren;
 Window r;
 unsigned int i;

 if (!(XQueryTree (dis, w, &dummy, &dummy, &children, &nChildren))) {
  return None;
 }

 for (i = 0; i < nChildren; ++i) {
  r = windowMatchTitle (dis, children[i], titleToFind);

  if (None != r) {
   XFree (children);
   return r;
  }
 
  Tcl_GetTime (&thisTime);
  if ((thisTime.sec - startTime->sec) >= SEARCH_N_SECONDS_FOR_WINDOW) {
   break;
  }

  r = searchWindowTree (dis, children[i], titleToFind, startTime);

  if (None != r) {
   XFree (children);
   return r;
  }
 }

 if (children) {
  XFree (children);
 }

 return None;
}


static Window findWindowViaTitle (Display *dis, Window parent, char *titleToFind) {
 Tcl_Time startTime, thisTime;
 Window r;

 Tcl_GetTime (&startTime);
 
 do {
  r = searchWindowTree (dis, parent, titleToFind, &startTime);

  if (None != r) {
   return r;
  }

  Tcl_GetTime (&thisTime);

 } while ((thisTime.sec - startTime.sec) < SEARCH_N_SECONDS_FOR_WINDOW);

 return None;
}

int TkXext_FindWindow TCL_OBJ_CMD_ARGS {
 Tcl_Obj *winObj;
 Tk_Window tkwin;
 Display *dis;
 char buf[30];

 if (2 != objc) {
  Tcl_WrongNumArgs (interp, 1, objv, "window-title");
  return TCL_ERROR;
 }

 tkwin = Tk_MainWindow (interp);
 Tk_MakeWindowExist (tkwin);
 dis = Tk_Display (tkwin);

 /*An XID is usually an unsigned long, so we can't use Tcl_NewLongObj*/
 snprintf (buf, sizeof(buf), "%lx", (unsigned long) 
  findWindowViaTitle (dis, RootWindow (dis, Tk_ScreenNumber(tkwin)),
  Tcl_GetString (objv[1])));
  
 winObj = Tcl_NewStringObj (buf, -1);
 Tcl_IncrRefCount (winObj);
 Tcl_SetObjResult (interp, winObj);
 Tcl_DecrRefCount (winObj);

 return TCL_OK;
}
